@@ -21,13 +21,14 @@ import mailbox
 import pwd
 
 from pathlib import Path
-from tempfile import mkstemp
+from tempfile import mkstemp, TemporaryDirectory
+from contextlib import contextmanager
 
 from email import charset
 charset.add_charset('utf-8', None)
 emlpolicy = email.policy.EmailPolicy(utf8=True, cte_type='8bit', max_line_length=None)
 
-__VERSION__ = '0.4.2'
+__VERSION__ = '0.5.0'
 ATTESTATION_FORMAT_VER = '0.1'
 
 logger = logging.getLogger('b4')
@@ -275,8 +276,11 @@ class LoreMailbox:
                 if pmsg is None:
                     # Can't find the message we're replying to here
                     continue
-            else:
+            elif fmsg.in_reply_to in self.msgid_map:
                 pmsg = self.msgid_map[fmsg.in_reply_to]
+            else:
+                logger.debug('  missing message, skipping: %s', fmsg.in_reply_to)
+                continue
 
             trailers, mismatches = fmsg.get_trailers(sloppy=sloppytrailers)
             for tname, tvalue in mismatches:
@@ -331,19 +335,19 @@ class LoreMailbox:
         logger.debug('Looking at: %s', lmsg.full_subject)
         self.msgid_map[lmsg.msgid] = lmsg
 
-        if lmsg.counter == 0 and lmsg.has_diffstat:
-            # Cover letter
-            # Add it to covers -- we'll deal with them later
-            logger.debug('  adding as v%s cover letter', lmsg.revision)
-            self.covers[lmsg.revision] = lmsg
-            return
-
         if lmsg.reply:
             # We'll figure out where this belongs later
             logger.debug('  adding to followups')
             self.followups.append(lmsg)
             return
 
+        if lmsg.counter == 0 and (not lmsg.counters_inferred or lmsg.has_diffstat):
+            # Cover letter
+            # Add it to covers -- we'll deal with them later
+            logger.debug('  adding as v%s cover letter', lmsg.revision)
+            self.covers[lmsg.revision] = lmsg
+            return
+
         if re.search(r'^Comment: att-fmt-ver:', lmsg.body, re.I | re.M):
             logger.debug('Found attestation message')
             LoreAttestationDocument.load_from_string(lmsg.msgid, lmsg.body)
@@ -398,16 +402,11 @@ class LoreSeries:
         self.trailer_mismatches = set()
         self.complete = False
         self.has_cover = False
+        self.subject = '(untitled)'
 
     def __repr__(self):
         out = list()
-        if self.has_cover:
-            out.append('- Series: [v%s] %s' % (self.revision, self.patches[0].subject))
-        elif self.patches[1] is not None:
-            out.append('- Series: [v%s] %s' % (self.revision, self.patches[1].subject))
-        else:
-            out.append('- Series: [v%s] (untitled)' % self.revision)
-
+        out.append('- Series: [v%s] %s' % (self.revision, self.subject))
         out.append('  revision: %s' % self.revision)
         out.append('  expected: %s' % self.expected)
         out.append('  complete: %s' % self.complete)
@@ -439,6 +438,10 @@ class LoreSeries:
         else:
             self.patches[lmsg.counter] = lmsg
         self.complete = not (None in self.patches[1:])
+        if self.patches[0] is not None:
+            self.subject = self.patches[0].subject
+        elif self.patches[1] is not None:
+            self.subject = self.patches[1].subject
 
     def get_slug(self, extended=False):
         # Find the first non-None entry
@@ -465,7 +468,7 @@ class LoreSeries:
 
         return slug
 
-    def save_am_mbox(self, mbx, noaddtrailers, covertrailers, trailer_order=None, addmysob=False,
+    def save_am_mbox(self, mbx, noaddtrailers=False, covertrailers=False, trailer_order=None, addmysob=False,
                      addlink=False, linkmask=None, cherrypick=None):
 
         usercfg = get_user_config()
@@ -584,6 +587,163 @@ class LoreSeries:
 
         return mbx
 
+    def check_applies_clean(self, gitdir, when=None):
+        # Go through indexes and see if this series should apply cleanly
+        mismatches = 0
+        seenfiles = set()
+        for lmsg in self.patches[1:]:
+            if lmsg is None or lmsg.blob_indexes is None:
+                continue
+            for fn, bh in lmsg.blob_indexes:
+                if fn in seenfiles:
+                    # if we have seen this file once already, then it's a repeat patch
+                    # and it's no longer going to match current hash
+                    continue
+                seenfiles.add(fn)
+                if set(bh) == {'0'}:
+                    # New file, will for sure apply clean
+                    continue
+                fullpath = os.path.join(gitdir, fn)
+                if when is None:
+                    if not os.path.exists(fullpath):
+                        mismatches += 1
+                        continue
+                    cmdargs = ['hash-object', fullpath]
+                    ecode, out = git_run_command(None, cmdargs)
+                else:
+                    gitdir = os.path.join(gitdir, '.git')
+                    logger.debug('Checking hash on %s:%s', when, fn)
+                    # XXX: We should probably pipe the two commands instead of reading into memory,
+                    #      so something to consider for the future
+                    ecode, out = git_run_command(gitdir, ['show', f'{when}:{fn}'])
+                    if ecode > 0:
+                        # Couldn't get this file, continue
+                        logger.debug('Could not look up %s:%s', when, fn)
+                        mismatches += 1
+                        continue
+                    cmdargs = ['hash-object', '--stdin']
+                    ecode, out = git_run_command(None, cmdargs, stdin=out.encode())
+                if ecode == 0:
+                    if out.find(bh) != 0:
+                        logger.debug('%s hash: %s (expected: %s)', fn, out.strip(), bh)
+                        mismatches += 1
+                    else:
+                        logger.debug('%s hash: matched', fn)
+
+        return len(seenfiles), mismatches
+
+    def make_fake_am_range(self, gitdir):
+        start_commit = end_commit = None
+        # Do we have it in cache already?
+        cachedir = get_cache_dir()
+        # Use the msgid of the first non-None patch in the series
+        msgid = None
+        for lmsg in self.patches:
+            if lmsg is not None:
+                msgid = lmsg.msgid
+                break
+        if msgid is None:
+            logger.critical('Cannot operate on an empty series')
+            return None, None
+        cachefile = os.path.join(cachedir, '%s.fakeam' % urllib.parse.quote_plus(msgid))
+        if os.path.exists(cachefile):
+            stalecache = False
+            with open(cachefile, 'r') as fh:
+                cachedata = fh.read()
+                chunks = cachedata.strip().split()
+                if len(chunks) == 2:
+                    start_commit, end_commit = chunks
+                else:
+                    stalecache = True
+            if start_commit is not None and end_commit is not None:
+                # Make sure they are still there
+                ecode, out = git_run_command(gitdir, ['cat-file', '-e', start_commit])
+                if ecode > 0:
+                    stalecache = True
+                else:
+                    ecode, out = git_run_command(gitdir, ['cat-file', '-e', end_commit])
+                    if ecode > 0:
+                        stalecache = True
+                    else:
+                        logger.debug('Using previously generated range')
+                        return start_commit, end_commit
+
+            if stalecache:
+                logger.debug('Stale cache for [v%s] %s', self.revision, self.subject)
+                os.unlink(cachefile)
+
+        logger.info('Preparing fake-am for v%s: %s', self.revision, self.subject)
+        with git_temp_worktree(gitdir):
+            # We are in a temporary chdir at this time, so writing to a known file should be safe
+            mbxf = '.__git-am__'
+            mbx = mailbox.mbox(mbxf)
+            # Logic largely borrowed from gj_tools
+            seenfiles = set()
+            for lmsg in self.patches[1:]:
+                if lmsg is None:
+                    logger.critical('ERROR: v%s series incomplete; unable to create a fake-am range', self.revision)
+                    return None, None
+                logger.debug('Looking at %s', lmsg.full_subject)
+                lmsg.load_hashes()
+                if not len(lmsg.blob_indexes):
+                    logger.critical('ERROR: some patches do not have indexes')
+                    logger.critical('       unable to create a fake-am range')
+                    return None, None
+                for fn, fi in lmsg.blob_indexes:
+                    if fn in seenfiles:
+                        # We already processed this file, so this blob won't match
+                        continue
+                    seenfiles.add(fn)
+                    if set(fi) == {'0'}:
+                        # New file creation, nothing to do here
+                        logger.debug('  New file: %s', fn)
+                        continue
+                    # Try to grab full ref_id of this hash
+                    ecode, out = git_run_command(gitdir, ['rev-parse', fi])
+                    if ecode > 0:
+                        logger.critical('  ERROR: Could not find matching blob for %s (%s)', fn, fi)
+                        logger.critical('         If you know on which tree this patchset is based,')
+                        logger.critical('         add it as a remote and perform "git remote update"')
+                        logger.critical('         in order to fetch the missing objects.')
+                        return None, None
+                    logger.debug('  Found matching blob for: %s', fn)
+                    fullref = out.strip()
+                    gitargs = ['update-index', '--add', '--cacheinfo', f'0644,{fullref},{fn}']
+                    ecode, out = git_run_command(None, gitargs)
+                    if ecode > 0:
+                        logger.critical('  ERROR: Could not run update-index for %s (%s)', fn, fullref)
+                        return None, None
+                mbx.add(lmsg.msg.as_string(policy=emlpolicy).encode('utf-8'))
+
+            mbx.close()
+            ecode, out = git_run_command(None, ['write-tree'])
+            if ecode > 0:
+                logger.critical('ERROR: Could not write fake-am tree')
+                return None, None
+            treeid = out.strip()
+            # At this point we have a worktree with files that should cleanly receive a git am
+            gitargs = ['commit-tree', treeid + '^{tree}', '-F', '-']
+            ecode, out = git_run_command(None, gitargs, stdin='Initial fake commit'.encode('utf-8'))
+            if ecode > 0:
+                logger.critical('ERROR: Could not commit-tree')
+                return None, None
+            start_commit = out.strip()
+            git_run_command(None, ['reset', '--hard', start_commit])
+            ecode, out = git_run_command(None, ['am', mbxf])
+            if ecode > 0:
+                logger.critical('ERROR: Could not fake-am version %s', self.revision)
+                return None, None
+            ecode, out = git_run_command(None, ['rev-parse', 'HEAD'])
+            end_commit = out.strip()
+            logger.info('  range: %.12s..%.12s', start_commit, end_commit)
+
+        with open(cachefile, 'w') as fh:
+            logger.debug('Saving into cache: %s', cachefile)
+            logger.debug('    %s..%s', start_commit, end_commit)
+            fh.write(f'{start_commit} {end_commit}\n')
+
+        return start_commit, end_commit
+
     def save_cover(self, outfile):
         cover_msg = self.patches[0].get_am_message(add_trailers=False, trailer_order=None)
         with open(outfile, 'w') as fh:
@@ -630,6 +790,8 @@ class LoreMessage:
         self.attestation = None
         # Patchwork hash
         self.pwhash = None
+        # Blob indexes
+        self.blob_indexes = None
 
         self.msgid = LoreMessage.get_clean_msgid(self.msg)
         self.lsubject = LoreSubject(msg['Subject'])
@@ -679,7 +841,13 @@ class LoreMessage:
             pcharset = part.get_content_charset()
             if not pcharset:
                 pcharset = mcharset
-            payload = payload.decode(pcharset, errors='replace')
+            try:
+                payload = payload.decode(pcharset, errors='replace')
+            except LookupError:
+                # what kind of encoding is that?
+                # Whatever, we'll use utf-8 and hope for the best
+                payload = payload.decode('utf-8', errors='replace')
+                part.set_param('charset', 'utf-8')
             if self.body is None:
                 self.body = payload
                 continue
@@ -870,6 +1038,22 @@ class LoreMessage:
 
         return hashed.hexdigest()
 
+    @staticmethod
+    def get_indexes(diff):
+        indexes = set()
+        curfile = None
+        for line in diff.split('\n'):
+            if line.find('diff ') != 0 and line.find('index ') != 0:
+                continue
+            matches = re.search(r'^diff\s+--git\s+\w/(.*)\s+\w/(.*)$', line)
+            if matches and matches.groups()[0] == matches.groups()[1]:
+                curfile = matches.groups()[0]
+                continue
+            matches = re.search(r'^index\s+([0-9a-f]+)\.\.[0-9a-f]+.*$', line)
+            if matches and curfile is not None:
+                indexes.add((curfile, matches.groups()[0]))
+        return indexes
+
     @staticmethod
     def get_clean_diff(diff):
         diff = diff.replace('\r', '')
@@ -944,19 +1128,6 @@ class LoreMessage:
             buflines.append(line)
         return difflines
 
-    @staticmethod
-    def get_patch_hash(diff):
-        # The aim is to represent the patch as if you did the following:
-        # git diff HEAD~.. | dos2unix | sha256sum
-        #
-        # This subroutine removes anything at the beginning of diff data, like
-        # diffstat or any other auxiliary data, and anything trailing at the end
-        #
-        diff = LoreMessage.get_clean_diff(diff)
-        phasher = hashlib.sha256()
-        phasher.update(diff.encode('utf-8'))
-        return phasher.hexdigest()
-
     def load_hashes(self):
         if self.attestation is not None:
             return
@@ -989,8 +1160,13 @@ class LoreMessage:
         with open(patch_out[1], 'r') as pfh:
             patch = pfh.read()
             if len(patch.strip()):
-                p = LoreMessage.get_patch_hash(patch)
+                diff = LoreMessage.get_clean_diff(patch)
+                phasher = hashlib.sha256()
+                phasher.update(diff.encode('utf-8'))
+                p = phasher.hexdigest()
                 self.pwhash = LoreMessage.get_patchwork_hash(patch)
+                # Load the indexes, if we have them
+                self.blob_indexes = LoreMessage.get_indexes(diff)
         os.unlink(patch_out[1])
 
         if i and m and p:
@@ -1063,49 +1239,47 @@ class LoreMessage:
         return githeaders, message, trailers, basement, signature
 
     def fix_trailers(self, trailer_order=None):
-        bodylines = self.body.split('\n')
-        # Get existing trailers
-        # 1. Find the first ---
-        # 2. Go backwards and grab everything matching ^[\w-]+:\s.*$ until a blank line
-        fixlines = list()
-        trailersdone = False
-        for line in bodylines:
-            if trailersdone:
-                fixlines.append(line)
-                continue
+        bheaders, message, btrailers, basement, signature = LoreMessage.get_body_parts(self.body)
+        # Now we add mix-in trailers
+        trailers = btrailers + list(self.followup_trailers)
+        fixtrailers = list()
+        if trailer_order is None:
+            trailer_order = DEFAULT_TRAILER_ORDER
+        for trailermatch in trailer_order:
+            for trailer in trailers:
+                if trailer in fixtrailers:
+                    # Dupe
+                    continue
+                if fnmatch.fnmatch(trailer[0].lower(), trailermatch.strip()):
+                    fixtrailers.append(trailer)
+                    if trailer not in btrailers:
+                        logger.info('    + %s: %s' % trailer)
+                    else:
+                        logger.debug('    . %s: %s' % trailer)
 
-            if line.strip() == '---':
-                # Start going backwards in fixlines
-                btrailers = list()
-                for rline in reversed(fixlines):
-                    if not len(rline.strip()):
-                        break
-                    matches = re.search(r'^([\w-]+):\s+(.*)', rline)
-                    if not matches:
-                        break
-                    fixlines.pop()
-                    btrailers.append(matches.groups())
+        # Reconstitute the message
+        if bheaders:
+            self.body = '\n'.join('%s: %s' % h for h in bheaders)
+            self.body += '\n\n'
+        else:
+            self.body = ''
 
-                # Now we add mix-in trailers
-                btrailers.reverse()
-                trailers = btrailers + list(self.followup_trailers)
-                added = list()
-                if trailer_order is None:
-                    trailer_order = DEFAULT_TRAILER_ORDER
-                for trailermatch in trailer_order:
-                    for trailer in trailers:
-                        if trailer in added:
-                            continue
-                        if fnmatch.fnmatch(trailer[0].lower(), trailermatch.strip()):
-                            fixlines.append('%s: %s' % trailer)
-                            if trailer not in btrailers:
-                                logger.info('    + %s: %s' % trailer)
-                            else:
-                                logger.debug('    . %s: %s' % trailer)
-                            added.append(trailer)
-                trailersdone = True
-            fixlines.append(line)
-        self.body = '\n'.join(fixlines)
+        if len(message):
+            self.body += message + '\n'
+            if len(fixtrailers):
+                self.body += '\n'
+
+        if len(fixtrailers):
+            self.body += '\n'.join('%s: %s' % t for t in fixtrailers)
+            self.body += '\n'
+        if len(basement):
+            self.body += '---\n'
+            self.body += basement
+            self.body += '\n'
+        if len(signature):
+            self.body += '-- \n'
+            self.body += signature
+            self.body += '\n'
 
     def get_am_message(self, add_trailers=True, trailer_order=None):
         if add_trailers:
@@ -1575,6 +1749,33 @@ def git_get_command_lines(gitdir, args):
     return lines
 
 
+@contextmanager
+def git_temp_worktree(gitdir=None):
+    """Context manager that creates a temporary work tree and chdirs into it. The
+    worktree is deleted when the contex manager is closed. Taken from gj_tools."""
+    dfn = None
+    try:
+        with TemporaryDirectory() as dfn:
+            git_run_command(gitdir, ['worktree', 'add', '--detach', '--no-checkout', dfn])
+            with in_directory(dfn):
+                yield
+    finally:
+        if dfn is not None:
+            git_run_command(gitdir, ['worktree', 'remove', dfn])
+
+
+@contextmanager
+def in_directory(dirname):
+    """Context manager that chdirs into a directory and restores the original
+    directory when closed. Taken from gj_tools."""
+    cdir = os.getcwd()
+    try:
+        os.chdir(dirname)
+        yield True
+    finally:
+        os.chdir(cdir)
+
+
 def get_config_from_git(regexp, defaults=None):
     args = ['config', '-z', '--get-regexp', regexp]
     ecode, out = git_run_command(None, args)
@@ -1755,7 +1956,13 @@ def save_strict_thread(in_mbx, out_mbx, msgid):
         logger.info('Reduced thread to strict matches only (%s->%s)', len(in_mbx), len(out_mbx))
 
 
-def get_pi_thread_by_url(t_mbx_url, savefile):
+def get_pi_thread_by_url(t_mbx_url, savefile, nocache=False):
+    cachedir = get_cache_dir()
+    cachefile = os.path.join(cachedir, '%s.pi.mbx' % urllib.parse.quote_plus(t_mbx_url))
+    if os.path.exists(cachefile) and not nocache:
+        logger.debug('Using cached copy: %s', cachefile)
+        shutil.copyfile(cachefile, savefile)
+        return savefile
     session = get_requests_session()
     resp = session.get(t_mbx_url)
     if resp.status_code != 200:
@@ -1769,22 +1976,13 @@ def get_pi_thread_by_url(t_mbx_url, savefile):
     with open(savefile, 'wb') as fh:
         logger.debug('Saving %s', savefile)
         fh.write(t_mbox)
+    shutil.copyfile(savefile, cachefile)
     return savefile
 
 
 def get_pi_thread_by_msgid(msgid, savefile, useproject=None, nocache=False):
     qmsgid = urllib.parse.quote_plus(msgid)
     config = get_main_config()
-    cachedir = get_cache_dir()
-    base = msgid
-    if useproject:
-        base = '%s-%s' % (useproject, msgid)
-    cachefile = os.path.join(cachedir, '%s.pi.mbx' % urllib.parse.quote_plus(base))
-    if os.path.exists(cachefile) and not nocache:
-        logger.debug('Using cached copy: %s', cachefile)
-        shutil.copyfile(cachefile, savefile)
-        return savefile
-
     # Grab the head from lore, to see where we are redirected
     midmask = config['midmask'] % qmsgid
     loc = urllib.parse.urlparse(midmask)
@@ -1805,7 +2003,7 @@ def get_pi_thread_by_msgid(msgid, savefile, useproject=None, nocache=False):
     logger.debug('t_mbx_url=%s', t_mbx_url)
 
     logger.critical('Grabbing thread from %s', projurl.split('://')[1])
-    in_mbxf = get_pi_thread_by_url(t_mbx_url, '%s-loose' % savefile)
+    in_mbxf = get_pi_thread_by_url(t_mbx_url, '%s-loose' % savefile, nocache=nocache)
     if not in_mbxf:
         return None
     in_mbx = mailbox.mbox(in_mbxf)
@@ -1814,7 +2012,6 @@ def get_pi_thread_by_msgid(msgid, savefile, useproject=None, nocache=False):
     in_mbx.close()
     out_mbx.close()
     os.unlink(in_mbxf)
-    shutil.copyfile(savefile, cachefile)
     return savefile
 
 
