@@ -98,8 +98,8 @@ def mbox_to_am(mboxfile, cmdargs):
         cherrypick = None
     logger.critical('Writing %s', am_filename)
     mbx = mailbox.mbox(am_filename)
-    am_mbx = lser.save_am_mbox(mbx, cmdargs.noaddtrailers, covertrailers,
-                               trailer_order=config['trailer-order'],
+    am_mbx = lser.save_am_mbox(mbx, noaddtrailers=cmdargs.noaddtrailers,
+                               covertrailers=covertrailers, trailer_order=config['trailer-order'],
                                addmysob=cmdargs.addmysob, addlink=cmdargs.addlink,
                                linkmask=config['linkmask'], cherrypick=cherrypick)
     logger.info('---')
@@ -123,6 +123,23 @@ def mbox_to_am(mboxfile, cmdargs):
             logger.critical('         From: %s <%s>', fname, femail)
         logger.critical('NOTE: Rerun with -S to apply them anyway')
 
+    topdir = None
+    # Are we in a git tree and if so, what is our toplevel?
+    gitargs = ['rev-parse', '--show-toplevel']
+    lines = b4.git_get_command_lines(None, gitargs)
+    if len(lines) == 1:
+        topdir = lines[0]
+
+    if cmdargs.threeway:
+        if not topdir:
+            logger.critical('WARNING: cannot prepare 3-way (not in a git dir)')
+        elif not lser.complete:
+            logger.critical('WARNING: cannot prepare 3-way (series incomplete)')
+        else:
+            rstart, rend = lser.make_fake_am_range(gitdir=None)
+            if rstart and rend:
+                logger.info('Prepared a fake commit range for 3-way merge (%.12s..%.12s)', rstart, rend)
+
     logger.critical('---')
     if not lser.complete:
         logger.critical('WARNING: Thread incomplete!')
@@ -164,8 +181,40 @@ def mbox_to_am(mboxfile, cmdargs):
         logger.critical('       git checkout -b %s %s', gitbranch, base_commit)
         logger.critical('       git am %s', am_filename)
     else:
-        logger.critical(' Base: not found, sorry')
-        logger.critical('       git checkout -b %s master', gitbranch)
+        cleanmsg = ''
+        if topdir is not None:
+            checked, mismatches = lser.check_applies_clean(topdir)
+            if mismatches == 0 and checked != mismatches:
+                cleanmsg = ' (applies clean to current tree)'
+            elif cmdargs.guessbase:
+                # Look at the last 10 tags and see if it applies cleanly to
+                # any of them. I'm not sure how useful this is, but I'm going
+                # to put it in for now and maybe remove later if it causes
+                # problems or slowness
+                if checked != mismatches:
+                    best_matches = mismatches
+                    cleanmsg = ' (best guess: current tree)'
+                else:
+                    best_matches = None
+                # sort the tags by authordate
+                gitargs = ['tag', '-l', '--sort=-taggerdate']
+                lines = b4.git_get_command_lines(None, gitargs)
+                if lines:
+                    # Check last 10 tags
+                    for tag in lines[:10]:
+                        logger.debug('Checking base-commit possibility for %s', tag)
+                        checked, mismatches = lser.check_applies_clean(topdir, tag)
+                        if mismatches == 0 and checked != mismatches:
+                            cleanmsg = ' (applies clean to: %s)' % tag
+                            break
+                        # did they all mismatch?
+                        if checked == mismatches:
+                            continue
+                        if best_matches is None or mismatches < best_matches:
+                            best_matches = mismatches
+                            cleanmsg = ' (best guess: %s)' % tag
+
+        logger.critical(' Base: not found%s', cleanmsg)
         logger.critical('       git am %s', am_filename)
 
     am_mbx.close()
@@ -274,7 +323,7 @@ def am_mbox_to_quilt(am_mbx, q_dirname):
             sfh.write('%s\n' % patch_filename)
 
 
-def get_newest_series(mboxfile):
+def get_extra_series(mboxfile, direction=1, wantvers=None, nocache=False):
     # Open the mbox and find the latest series mentioned in it
     mbx = mailbox.mbox(mboxfile)
     base_msg = None
@@ -288,18 +337,18 @@ def get_newest_series(mboxfile):
         # Ignore replies or counters above 1
         if lsub.reply or lsub.counter > 1:
             continue
-        if latest_revision is None or lsub.revision > latest_revision:
-            # New revision
+        if base_msg is not None:
+            logger.debug('Current base_msg: %s', base_msg['Subject'])
+        logger.debug('Checking the subject on %s', lsub.full_subject)
+        if latest_revision is None or lsub.revision >= latest_revision:
             latest_revision = lsub.revision
-            if lsub.counter == 0:
+            if lsub.counter == 0 and not lsub.counters_inferred:
                 # And a cover letter, nice. This is the easy case
                 base_msg = msg
                 seen_covers.append(latest_revision)
-                continue
-            if lsub.counter == 1:
-                if latest_revision not in seen_covers:
-                    # A patch/series without a cover letter
-                    base_msg = msg
+            elif lsub.counter == 1 and latest_revision not in seen_covers:
+                # A patch/series without a cover letter
+                base_msg = msg
 
     # Get subject info from base_msg again
     lsub = b4.LoreSubject(base_msg['Subject'])
@@ -307,14 +356,27 @@ def get_newest_series(mboxfile):
         logger.debug('Not checking for new revisions: no prefixes on the cover letter.')
         mbx.close()
         return
+    if direction < 0 and latest_revision <= 1:
+        logger.debug('This is the latest version of the series')
+        mbx.close()
+        return
+    if direction < 0 and wantvers is None:
+        wantvers = [latest_revision - 1]
+
     base_msgid = b4.LoreMessage.get_clean_msgid(base_msg)
     fromeml = email.utils.getaddresses(base_msg.get_all('from', []))[0][1]
     msgdate = email.utils.parsedate_tz(str(base_msg['Date']))
     startdate = time.strftime('%Y%m%d', msgdate[:9])
     listarc = base_msg.get_all('List-Archive')[-1].strip('<>')
-    q = 's:"%s" AND f:"%s" AND d:%s..' % (lsub.subject.replace('"', ''), fromeml, startdate)
-    queryurl = '%s?%s' % (listarc, urllib.parse.urlencode({'q': q, 'x': 'A', 'o': '-1'}))
-    logger.critical('Checking for newer revisions on %s', listarc)
+    if direction > 0:
+        q = 's:"%s" AND f:"%s" AND d:%s..' % (lsub.subject.replace('"', ''), fromeml, startdate)
+        queryurl = '%s?%s' % (listarc, urllib.parse.urlencode({'q': q, 'x': 'A', 'o': '-1'}))
+        logger.critical('Checking for newer revisions on %s', listarc)
+    else:
+        q = 's:"%s" AND f:"%s" AND d:..%s' % (lsub.subject.replace('"', ''), fromeml, startdate)
+        queryurl = '%s?%s' % (listarc, urllib.parse.urlencode({'q': q, 'x': 'A', 'o': '1'}))
+        logger.critical('Checking for older revisions on %s', listarc)
+
     logger.debug('Query URL: %s', queryurl)
     session = b4.get_requests_session()
     resp = session.get(queryurl)
@@ -322,7 +384,7 @@ def get_newest_series(mboxfile):
     try:
         tree = xml.etree.ElementTree.fromstring(resp.content)
     except xml.etree.ElementTree.ParseError as ex:
-        logger.debug('Unable to parse results, ignoring', ex)
+        logger.debug('Unable to parse results, ignoring: %s', ex)
         resp.close()
         mbx.close()
         return
@@ -337,24 +399,36 @@ def get_newest_series(mboxfile):
             logger.debug('Ignoring result (not interesting): %s', title)
             continue
         link = entry.find('atom:link', ns).get('href')
-        if lsub.revision < latest_revision:
+        if direction > 0 and lsub.revision <= latest_revision:
             logger.debug('Ignoring result (not new revision): %s', title)
             continue
+        elif direction < 0 and lsub.revision >= latest_revision:
+            logger.debug('Ignoring result (not old revision): %s', title)
+            continue
+        elif direction < 0 and lsub.revision not in wantvers:
+            logger.debug('Ignoring result (not revision we want): %s', title)
+            continue
         if link.find('/%s/' % base_msgid) > 0:
             logger.debug('Ignoring result (same thread as ours):%s', title)
             continue
         if lsub.revision == 1 and lsub.revision == latest_revision:
             # Someone sent a separate message with an identical title but no new vX in the subject line
-            # It's *probably* a new revision.
-            logger.debug('Likely a new revision: %s', title)
-        elif lsub.revision > latest_revision:
+            if direction > 0:
+                # It's *probably* a new revision.
+                logger.debug('Likely a new revision: %s', title)
+            else:
+                # It's *probably* an older revision.
+                logger.debug('Likely an older revision: %s', title)
+        elif direction > 0 and lsub.revision > latest_revision:
             logger.debug('Definitely a new revision [v%s]: %s', lsub.revision, title)
+        elif direction < 0 and lsub.revision < latest_revision:
+            logger.debug('Definitely an older revision [v%s]: %s', lsub.revision, title)
         else:
             logger.debug('No idea what this is: %s', title)
             continue
         t_mbx_url = '%st.mbox.gz' % link
         savefile = mkstemp('b4-get')[1]
-        nt_mboxfile = b4.get_pi_thread_by_url(t_mbx_url, savefile)
+        nt_mboxfile = b4.get_pi_thread_by_url(t_mbx_url, savefile, nocache=nocache)
         nt_mbx = mailbox.mbox(nt_mboxfile)
         # Append all of these to the existing mailbox
         new_adds = 0
@@ -409,7 +483,7 @@ def main(cmdargs):
             sys.exit(1)
 
     if threadmbox and cmdargs.checknewer:
-        get_newest_series(threadmbox)
+        get_extra_series(threadmbox, direction=1)
 
     if cmdargs.subcmd == 'am':
         mbox_to_am(threadmbox, cmdargs)
